@@ -81,6 +81,28 @@ VANISHED_AGE = (60 * 60 * 24 * 30)
 # Kerberos authentication state constants
 (GSS_STATE_STEP, GSS_STATE_WRAP) = (0, 1)
 
+#
+# Bugfix classes
+#
+
+if sys.version_info < (2, 5, 3):
+    # A serious imaplib bug (http://bugs.python.org/issue1389051) was
+    # fixed in 2.5.3.  Earlier Python releases need a work-around.
+    orig_imap4_ssl = imaplib.IMAP4_SSL
+    class IMAP4_SSL(orig_imap4_ssl):
+        def read(self, size):
+            """Read 'size' bytes from remote."""
+            # sslobj.read() sometimes returns < size bytes
+            chunks = []
+            read = 0
+            while read < size:
+                data = self.sslobj.read(min(size-read, 16384))
+                read += len(data)
+                chunks.append(data)
+            return ''.join(chunks)
+    
+    imaplib.IMAP4_SSL = IMAP4_SSL
+
 #
 # Mix-in classes
 #
@@ -93,6 +115,7 @@ class POP3initMixIn(object):
         self.log.trace()
         try:
             self.conn = poplib.POP3(self.conf['server'], self.conf['port'])
+            self.setup_received(self.conn.sock)
         except poplib.error_proto, o:
             raise getmailOperationError('POP error (%s)' % o)
         except socket.timeout:
@@ -136,6 +159,7 @@ class Py24POP3SSLinitMixIn(object):
                                + os.linesep)
                 self.conn = poplib.POP3_SSL(self.conf['server'],
                                             self.conf['port'])
+            self.setup_received(self.conn.sock)
         except poplib.error_proto, o:
             raise getmailOperationError('POP error (%s)' % o)
         except socket.timeout:
@@ -180,6 +204,7 @@ class Py23POP3SSLinitMixIn(object):
                     + os.linesep
                 )
                 self.conn = POP3SSL(self.conf['server'], self.conf['port'])
+            self.setup_received(self.conn.rawsock)
         except poplib.error_proto, o:
             raise getmailOperationError('POP error (%s)' % o)
         except socket.timeout:
@@ -202,6 +227,7 @@ class IMAPinitMixIn(object):
         self.log.trace()
         try:
             self.conn = imaplib.IMAP4(self.conf['server'], self.conf['port'])
+            self.setup_received(self.conn.sock)
         except imaplib.IMAP4.error, o:
             raise getmailOperationError('IMAP error (%s)' % o)
         except socket.timeout:
@@ -242,6 +268,7 @@ class IMAPSSLinitMixIn(object):
                 )
                 self.conn = imaplib.IMAP4_SSL(self.conf['server'],
                                               self.conf['port'])
+            self.setup_received(self.conn.sock)
         except imaplib.IMAP4.error, o:
             raise getmailOperationError('IMAP error (%s)' % o)
         except socket.timeout:
@@ -329,6 +356,21 @@ class RetrieverSkeleton(ConfigurableBase):
         self.gotmsglist = False
         ConfigurableBase.__init__(self, **args)
 
+    def setup_received(self, sock):
+        serveraddr = sock.getpeername()
+        if len(serveraddr) == 2:
+            # IPv4
+            self.remoteaddr = '%s:%s' % serveraddr
+        elif len(serveraddr) == 4:
+            # IPv6
+            self.remoteaddr = '[%s]:%s' % serveraddr[:2]
+        else:
+            # Shouldn't happen
+            self.log.warn('unexpected peer address format %s', str(serveraddr))
+            self.remoteaddr = str(serveraddr)
+        self.received_from = '%s (%s)' % (self.conf['server'], 
+                                          self.remoteaddr)
+
     def __del__(self):
         self.log.trace()
         self.write_oldmailfile()
@@ -420,10 +462,7 @@ class RetrieverSkeleton(ConfigurableBase):
         self.oldmail_filename = os.path.join(self.conf['getmaildir'], 
                                              oldmail_filename)
         self._read_oldmailfile()
-        self.received_from = '%s (%s)' % (
-            self.conf['server'],
-            socket.getaddrinfo(self.conf['server'], None)[0][4][0]
-        )
+        self.received_from = None
         self.__initialized = True
 
     def delivered(self, msgid):
@@ -724,8 +763,8 @@ class IMAPRetrieverBase(RetrieverSkeleton):
             raise getmailOperationError('IMAP error (%s)' % o)
         if result != 'OK':
             raise getmailOperationError(
-                'IMAP error (command %s returned %s)'
-                % ('%s %s' % (cmd, args), result)
+                'IMAP error (command %s returned %s %s)'
+                % ('%s %s' % (cmd, args), result, resplist)
             )
         if cmd.lower().startswith('login'):
             self.log.debug('login command response %s' % resplist + os.linesep)
@@ -745,8 +784,8 @@ class IMAPRetrieverBase(RetrieverSkeleton):
             raise getmailOperationError('IMAP error (%s)' % o)
         if result != 'OK':
             raise getmailOperationError(
-                'IMAP error (command %s returned %s)'
-                % ('%s %s' % (cmd, args), result)
+                'IMAP error (command %s returned %s %s)'
+                % ('%s %s' % (cmd, args), result, resplist)
             )
         self.log.debug('command uid %s response %s'
                        % ('%s %s' % (cmd, args), resplist) + os.linesep)
@@ -886,7 +925,14 @@ class IMAPRetrieverBase(RetrieverSkeleton):
             # (virus, spam, trojan), it can completely fail to return the
             # message when requested.
             try:
-                msg = Message(fromstring=response[0][1])
+                try:
+                    sbody = response[0][1]
+                except Exception, o:
+                    sbody = None
+                if not sbody:
+                    self.log.error('bad message from server!')
+                    sbody = str(response)
+                msg = Message(fromstring=sbody)
             except TypeError, o:
                 # response[0] is None instead of a message tuple
                 raise getmailRetrievalError('failed to retrieve msgid %s' 
